;********************************
;                               ;
;   Copyright Micro AMPS Ltd    ;
;   & Philips Components        ;
;   Dec 1990                    ;
;                               ;
;********************************

; Read data under interrupt from an IBM keyboard
; Hardware resources:
;    Kbd clock on interrupt INT0 P1.5
;    Kbd data on pin             P0.2


; This pogram reads keys in from the keyboard
; and translates them to ASCII



#include equates.51
#include kbd.h

IICADD      .equ 088h        ; our i2c slave address
MAXBYTES    .equ 1           ; max bytes to recv or trans

rcvdat      .equ 04h         ; I2C received data buffer
xmtdat      .equ 06h         ; I2C transmitter buffer

STACK       .equ 08h

flags       .equ 020h          ; byte used as flags
noack       .equ (flags-20h)   ; I2C flags.0, ...1, ...2, etc
recvd       .equ (flags-20h)+1 ;
sent_flag   .equ (flags-20h)+2 ;
i2c_busy    .equ (flags-20h)+3 ;

Cntrl       .equ (flags-20h)+8 ; Control key flag
Shift       .equ (flags-20h)+9 ; Shift   key flag

bitcnt      .equ flags+2
bytecnt     .equ flags+3

adrrcvd     .equ flags+4
rwflag      .equ (adrrcvd-20h)*8  ; adrrcvd.0

tick        .equ 025h        ; count 10mS ticks to give 1sec tick
i2ctime     .equ 027h        ; i2c timeout - used on slow i2c bus


NBits       .equ 29h         ; # bits read so far
NBytes      .equ NBits+1     ; # bytes in buffer
lastkey     .equ NBytes+1    ; last key was?
keytemp     .equ lastkey+1   ; used to build the key bit by bit
keybuff     .equ keytemp+1   ; store the chars here


INMAX       .equ 8           ; Size of keyboard buffer
KEYCLK      .equ p1.5        ; keyboard clock signal on ext int 0
KEYDAT      .equ 82h         ; keyboard data line

EDGEINT     .equ 08ah

;  reset and interrupt vectors.
;
        .org 0               ; reset vector
        ajmp start

        .org 0003h           ; external Interrupt 0
        ajmp kbd

        .org 0bh             ; counter/timer 0
        ajmp badint

        .org 013h            ; ext int 1
        ajmp badint

        .org 01bh            ; timer 1 - i2c timeout
        ajmp timerI

        .org 023h            ; i2c interupt
        ajmp i2cint


        .org  48h

done:
        ajmp $               ; main routine waiting for key presses

        .org  50h

start:
        mov r0,#0ffh         ; power supply settling time
        mov r1,#0ffh
        mov r2,#04h

dly1:   djnz r0,$
        djnz r1,dly1
        djnz r2,dly1

reset:
        mov p0,#0ffh
        mov p1,#0ffh
        mov p3,#0ffh

        mov sp,#STACK        ; initialise stack pointer
        setb EDGEINT         ; make ext int 0 edge activated

        clr Shift            ; clear keyboard shift flag

        mov r0,#NBits        ; clear the input buffers
        mov r1,#10h
        mov acc,#0

clrlp:  mov @r0,a            ; do the clearing
        inc r0
        djnz r1,clrlp

;       mov xmtdat,#'.'      ; transmitt buffer filled with
;       mov xmtdat+1,#0ffh   ; $ff when empty

        mov flags,#0
        mov i2ctime,#0

restart:

        mov  I2CFG,#80h+CTVAL ;enable slave functions
        mov  I2CON,#BIDLE    ;place in idle state

        mov ien0,#91h        ;enable external & IIC interrupts


;******  Main loop  ******

main:
        mov a,NBytes
        jz empty             ; if data in keybuff copy to i2c xmt buffer

        mov ien0,#0h         ;disable all interrupts temporarily
        mov xmtdat,keybuff
        mov NBytes,#0        ; clear keyboard buffer full flag
        mov ien0,#91h        ;enable external & IIC interrupts

empty:

        jnb recvd,notread    ; recvd flag tells 751 to clear I2C xmt buffer

        mov acc,xmtdat       ; when i2c master sucessfully reads the data
        orl a,#80h           ; from the 87c751, the master writes any data
        mov xmtdat,acc       ; back which will set the MSB of the data
                             ; buffer. This is necessary to syncronise
                             ; the two processors.
        clr recvd            ; reset i2c received flag

notread:

        mov a,lastkey        ; detect alt key for special functions
        clr c
        subb a,#11h
        jnz notalt

        nop                  ; alt code goes here

notalt:
        mov a,NBytes         ;
        clr c
        subb a,#INMAX        ; limit the input buffer to INMAX
        jc notdone

        nop                  ; if data is buffered then buffer exceed
                             ; code goes here

notdone:

        sjmp main            ; go back to start

;******  End of Main loop  ******


; ******   External int 0  ISR  ******
;
; keyboard interrupt service routine
;
; ************************************


kbd:
        push psw             ; save regs during ISR
        push acc

        mov acc,NBits        ; NBits = the bit number next expected
                             ; from the keyboard
        cjne a,#0,bit1_8     ; if not bit 0 then maybe bit 1 to 8



;******  Keyboard Bit 0  ******

bit0:                        ; discard bit 0 - Start bit
        sjmp bump


;****** Keyboard Bit 1-8  ******

bit1_8:
        cjne a,#9,$+3        ; CY flag is set if acc < 9
        jnc bit9

        mov c,KEYDAT         ; read data for keyboard data line
        mov a,keytemp        ; data arrives least sig bit 1st
        rr a                 ; hence old value is rotated and new
        mov a.7,c            ; bit is or'ed to the msb
        mov keytemp,acc
        sjmp bump


;******  Bit 9  ******

bit9:
        cjne a,#9,bit10

        sjmp bump            ; parity check code would go here



;****** Bit 10  ******

; The stop bit - Key Scan is now complete so convert to Ascii

bit10:
        mov acc,keytemp      ; get next key

        cjne a,#12h,notls    ; is it the left shift char?


;****** Left Shift has Been Pressed  ******

        mov acc,lastkey
        cjne a,#0f0h,makels  ; if last key was $f0 then shift is released

        clr Shift            ; Next keys will be unshifted
        mov lastkey,#12h     ; Copy left shift key to last key
        sjmp tidy

makels:                      
        setb Shift           ; Next Keys will be shifted
        mov lastkey,#12h     ; Copy left shift key to last key
        sjmp tidy

;****** End of Shift Routine  ******

notls:
;       mov acc,keytemp      ; get next key
        cjne a,#14h,notctrl  ; is it a control char?


;****** Control State  ******

        nop                  ; control state goes here
        sjmp tidy

;****** End of Control State  ******

notctrl:

        cjne a,#0f0h,notbreak ; is current key is $f0 then break code

;******  Key Break  ******

        mov lastkey,a        ; if so record break code in last key
        sjmp tidy            ; but don't store in the buffer

notbreak:

        mov acc,lastkey      ; if last key was $f0 then
        cjne a,#0f0h,not_f0  ; ignore the next scan code

        mov lastkey,#0       ; which is a break code
        sjmp tidy

not_f0:


;****** Normal Key Press  ******

#ifdef buffered

;******  Buffered Code  ******

; buffered code
        push 0               ; r0 will be used as an indirect pointer
                             ; so save it
        mov acc,#keybuff     ; copy data into keyboard
        add a,NBytes
        mov r0,a

        mov a,keytemp        ; get current key
        mov lastkey,a        ; & copy to lastkey

        push dph             ; dp required to point to translation
        push dpl             ; tables. Since in ISR save dp contents

        jb Shift,shifted     ; if in the unshifted state use the unshift table
        mov dptr,#unshift
        sjmp skip1

shifted:
        mov dptr,#shift      ; else use the shift table

skip1 :
        movc a,@a+dptr       ; translate the char in Acc to Ascii

        pop dpl              ; retsore the data pointer
        pop dph

        cjne a,#0,Not0       ; If data is zero discard

        ; sjmp NoSave        ; discard code goes here

Not0:
        mov r0,#keybuff      ; Save Ascii value into keyboard buffer
        mov @r0,a            ; buffered keyboard entry
        inc NBytes

NoSave
        pop 0              ; restore r0

;**** End of Buffered Code

#endif


#define unbuffered 1

#ifdef unbuffered

        mov a,keytemp        ; get current key
        mov lastkey,a        ; & copy to lastkey

        push dph             ; dp required to point to translation
        push dpl             ; tables. Since in ISR save dp contents

        jb Shift,shifted     ; if in the unshifted state use the unshift table
        mov dptr,#unshift
        sjmp skip1

shifted:
        mov dptr,#shift      ; else use the shift table

skip1 :
        movc a,@a+dptr       ; translate the char in Acc to Ascii

        pop dpl              ; retsore the data pointer
        pop dph

        cjne a,#0,Not0       ; If data is zero discard

        ; sjmp tidy          ; discard code goes here

Not0:
        mov keybuff,a        ; store in kbd buffer
        mov NBytes,#1        ; mark byte read

tidy:
        mov NBits,#0         ; clear flags ready for next keyboard char
        mov keytemp,#0
        sjmp intdone

;******  End of Keyboard Translation and Save  ******


;******  Normal unfinished key exit ******

bump:
        inc NBits            ; inc the number of bits read so far

intdone:
        pop acc
        pop psw
        reti

;****** End of Ext Int 0 ISR  ******




;******  I2C CODE SLAVE  ******

i2cint:                      ; i2c interrupt entry point
        setb i2c_busy        ; semaphore on xmtdata buffer

        push psw             ; save registers used in ISR
        push acc
        push 0               ; R0 no bank switching

        clr  EI2             ; make i2c ISR interruptable
        acall clrint         ; execute a reti

slave:
        mov i2ctime,#3       ; set up i2c timeout watchdog 30 mS

        mov I2CON,#BCARL+BCSTP+BCSTR+BCXA
                             ; clear start status

        jnb ATN,$            ; wait for next data bit
        mov bitcnt,#7

        acall recvb2         ; get remainder of slave address
        mov adrrcvd,a
        clr a.0              ; mask read/write bit to check address
        cjne a,#IICADD,goidle ; idle again if not for us

        jb rwflag,read       ; test for read or write



;****** I2C Receive Code  ******

        mov r0,#rcvdat       ; r0 points to data bufferer
        mov bytecnt,#MAXBYTES

rcvloop:
        acall sendack        ; acknowledge the address
        acall rcvbyte        ; wait for the next data byte
        jnb DRDY,exitwr      ; end of frame
        mov @r0,a
        inc r0
        djnz bytecnt,rcvloop

; no more room

        acall sendack        ; ack last byte
        acall rcvbyte        ; get but discard next one
        mov I2DAT,#80h       ; send neg ack
        jnb ATN,$            ; wait till gone
exitwr:

        setb recvd
        sjmp msgend

;******  End of Receive Routine  ******


;****** I2C Transmit Code  ******

read:
        mov r0,#xmtdat       ; r0 points to data buffer
        mov bytecnt,#MAXBYTES
        acall sendack        ; acknowledge address

txloop:
        mov a,@r0            ; get next data byte
        inc r0               ; bump buffer pointer
        acall xmitbyte       ; transmit the byte to the I2C
        jb noack,exitrd      ; if not acknowledged the exit
        djnz bytecnt,txloop

exitrd: sjmp msgend

;****** End of I2C transmit  ******


;****** Repeated start state  ******
msgend:
        jnb ATN,$            ; wait for stop or repeated start
        jb STR,slave         ; if repeat start do again

; stop so enter idle mode

goidle:
        mov i2ctime,#0       ; stop i2c timeout
        mov I2CON,#BCSTP+BCXA+BCDR+BCARL+BIDLE

        pop 0                ;restore state before i2c ISR
        pop acc
        pop psw

        setb EI2
        setb sent_flag       ;flag to say data has been sent
        clr i2c_busy         ;flag denotes exiting I2C routine

        ret


;******  General i2c routines  ******
                             
xmitbyte:                    ; I2C data transmit  ( enter with data in acc )
        mov bitcnt,#8

xmitbit:
        mov I2DAT,a
        rl a
        jnb ATN,$
        djnz bitcnt,xmitbit
        mov I2CON,#BCDR+BCXA ; switch to rcv mode
        jnb ATN,$            ; wait for ack
        mov flags,I2DAT      ; save ack bit
        ret



rdack:                       ; rdack   - receives data byte then sends ack
        acall rcvbyte        ; I2C data recv routine  - data returned in a

sendack:
        mov I2DAT,#0         ; I2C ack = data low when clock pulsed high
        jnb ATN,$
        ret

rcvbyte:                     ; rcvbyte receives byte - data returned in Acc
        mov bitcnt,#8

recvb2: clr a

rbit:   orl a,I2DAT
        rl a
        jnb ATN,$
        jnb DRDY,rbex        ; exit if not a data bit
        djnz bitcnt,rbit

        mov c,RDAT           ; get last bit - do not clear ATN
        rlc a                ; shift into byte

rbex:
        ret

;       IIC timer interrupt service
timerI:
        mov ien0,#0          ; break point address in ICE751
        setb CLRTI           ; clear the interrupt
fixup:
        mov I2CFG,#0         ; turn off i2c
        mov I2CON,#BCXA+BCDR+BCARL+BCSTR+BCSTP
                             ; reset i2c flags
        acall clrint
        ajmp reset           ; restart program


;******  call here to make code interruptable  ******

clrint:
        reti

;******  unused interrupts are vectored to here  ******
badint:
        reti

#include attable.h

        .end

